#! /usr/bin/env python3
import os
import sys
import time
import logging
import requests
from bs4 import BeautifulSoup
from typing import List, Dict, Any, Optional, Generator

from people_also_ask_it.tools import retryable
from people_also_ask_it.parser import (
    extract_related_questions,
    get_featured_snippet_parser,
)
from people_also_ask_it.exceptions import (
    GoogleSearchRequestFailedError,
    RelatedQuestionParserError,
    FeaturedSnippetParserError
)
from people_also_ask_it.tools import CallingSemaphore


URL = "https://www.google.it/search"
HEADERS = {
    'User-Agent': "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
    " AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/84.0.4147.135 Safari/537.36"
}
SESSION = requests.Session()
NB_TIMES_RETRY = 3
NB_REQUESTS_LIMIT = os.environ.get(
    "RELATED_QUESTION_NBREQUESTS_LIMIT", 25
)
NB_REQUESTS_DURATION_LIMIT = os.environ.get(
    "RELATED_QUESTION_DURATION_LIMIT", 90  # seconds
)
logging.basicConfig()
semaphore = CallingSemaphore(
    NB_REQUESTS_LIMIT, NB_REQUESTS_DURATION_LIMIT
)


@retryable(3)
def search(keyword: str) -> Optional[BeautifulSoup]:
    """return html parser of google search result"""
    params = {"q": keyword , "hl" : "it" }
    try:
        with semaphore:
            time.sleep(0.5)  # be nice with google :)
            response = SESSION.get(URL, params=params, headers=HEADERS)
    except Exception:
        raise GoogleSearchRequestFailedError(URL, keyword)
    if response.status_code != 200:
        raise GoogleSearchRequestFailedError(URL, keyword)
    return BeautifulSoup(response.text, "html.parser")


def _get_related_questions(text: str) -> List[str]:
    """
    return a list of questions related to text.
    These questions are from search result of text

    :param str text: text to search
    """
    document = search(text)
    if not document:
        return []
    try:
        return extract_related_questions(document)
    except Exception:
        raise RelatedQuestionParserError(text)


def generate_related_questions(text: str) -> Generator[str, None, None]:
    """
    generate the questions related to text,
    these quetions are found recursively

    :param str text: text to search
    """
    questions = set(_get_related_questions(text))
    searched_text = set(text)
    while questions:
        text = questions.pop()
        yield text
        searched_text.add(text)
        questions |= set(_get_related_questions(text))
        questions -= searched_text


def get_related_questions(text: str, max_nb_questions: Optional[int] = None):
    """
    return a number of questions related to text.
    These questions are found recursively.

    :param str text: text to search
    """
    if max_nb_questions is None:
        return _get_related_questions(text)
    nb_question_regenerated = 0
    questions = set()
    for question in generate_related_questions(text):
        if nb_question_regenerated > max_nb_questions:
            break
        questions.add(question)
        nb_question_regenerated += 1
    return list(questions)


def get_answer(question: str) -> Dict[str, Any]:
    """
    return a dictionary as answer for a question.

    :param str question: asked question
    """
    document = search(question)
    related_questions = extract_related_questions(document)
    featured_snippet = get_featured_snippet_parser(
            question, document)
    if not featured_snippet:
        res = dict(
            has_answer=False,
            question=question,
            related_questions=related_questions,
        )
    else:
        res = dict(
            has_answer=True,
            question=question,
            related_questions=related_questions,
        )
        try:
            res.update(featured_snippet.to_dict())
        except Exception:
            raise FeaturedSnippetParserError(question)
    return res


def generate_answer(text: str) -> Generator[dict, None, None]:
    """
    generate answers of questions related to text

    :param str text: text to search
    """
    answer = get_answer(text)
    questions = set(answer["related_questions"])
    searched_text = set(text)
    if answer["has_answer"]:
        yield answer
    while questions:
        text = questions.pop()
        answer = get_answer(text)
        if answer["has_answer"]:
            yield answer
        searched_text.add(text)
        questions |= set(get_answer(text)["related_questions"])
        questions -= searched_text


def get_simple_answer(question: str, depth: bool = False) -> str:
    """
    return a text as summary answer for the question

    :param str question: asked quetion
    :param bool depth: return the answer of first related question
        if no answer found for question
    """
    document = search(question)
    featured_snippet = get_featured_snippet_parser(
            question, document)
    if featured_snippet:
        return featured_snippet.response
    if depth:
        related_questions = get_related_questions(question)
        if not related_questions:
            return ""
        return get_simple_answer(related_questions[0])
    return ""


if __name__ == "__main__":
    from pprint import pprint as print
    print(get_answer(sys.argv[1]))



#! /usr/bin/env python3
import time
import traceback
from contextlib import ContextDecorator
from typing import Callable, List
from people_also_ask_it.exceptions import FeaturedSnippetParserError


def raise_featuredsnippetparsererror_if_failed(func):
    def wrapper(self: "SimpleFeaturedSnippetParser", *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception:
            traceback.print_exc()
            raise FeaturedSnippetParserError(self.text)
    return wrapper


def retryable(nb_times_retry):

    def decorator(func: Callable):

        def wrapper(*args, **kwargs):
            for _ in range(nb_times_retry-1):
                try:
                    return func(*args, **kwargs)
                except Exception:
                    pass
            return func(*args, **kwargs)

        return wrapper
    return decorator


def itemize(lines: List[str]) -> List[str]:
    return ["\t- " + line for line in lines]


def tabulate(header, table):
    length_columns = []
    if header:
        table = [header] + table
        length_columns = [len(str(e)) for e in header]
    for row in table:
        current_lengh = [len(str(e)) for e in row]
        length_columns = [
            max(i, j) for i, j in zip(length_columns, current_lengh)
        ]
    tabulated_rows = []
    for row in table:
        tabulated_rows.append("\t".join([
            str(e).rjust(length, " ") for e, length in zip(row, length_columns)
        ]))
    if header:
        tabulated_rows.insert(
            1,
            "\t".join(["-"*length for length in length_columns])
        )
    return "\n".join(tabulated_rows)


def remove_redundant(elements): return list(dict.fromkeys(elements))


class CallingSemaphore(ContextDecorator):

    def __init__(self, nb_call_times_limit, expired_time):
        self.nb_call_times_limit = nb_call_times_limit
        self.expired_time = expired_time
        self.called_timestamps = list()

    def __enter__(self):
        while len(self.called_timestamps) > self.nb_call_times_limit:
            now = time.time()
            self.called_timestamps = list(filter(
                lambda x: now - x < self.expired_time,
                self.called_timestamps
            ))
            time.sleep(0.5)
        self.called_timestamps.append(time.time())

    def __exit__(self, *exc):
        pass



#! /usr/bin/env python3
"""
Global realted-questions exception and warning classes.
"""


GITHUB_LINK = "https://github.com/lagranges/people_also_ask"


class RelatedQuestionError(Exception):
    """Base Related-Questions exception class."""

    def __init__(self, error):
        self.error = error

    def __unicode__(self):
        return (
            f'An unkown error occured: {self.error}.'
            f' Please report it on {GITHUB_LINK}.'
        )


class FeaturedSnippetParserError(RelatedQuestionError):
    """
    Exception raised when failed to get answer from
    search result page
    """

    def __init__(self, text):
        self.keyword = text

    def __unicode__(self):
        return (
            f"Cannot parse result page of '{self.text}'."
            f" It mays due to a format change of result page."
            f' Please report it on {GITHUB_LINK}.'
        )


class RelatedQuestionParserError(RelatedQuestionError):
    """
    Exception raised when failed to get related questions
    from search result page
    """

    def __init__(self, text):
        self.keyword = text

    def __unicode__(self):
        return (
            f"Cannot parse result page of '{self.text}'."
            f" It mays due to a format change of result page."
            f' Please report it on {GITHUB_LINK}.'
        )


class GoogleSearchRequestFailedError(RelatedQuestionError):
    """Exception raised when failed to request search on google"""

    def __init__(self, url, keyword):
        self.url = url
        self.keyword = keyword

    def __unicode__(self):
        return (
            f"Failed to requests {self.url}/{self.keyword}"
        )


class InvalidQuestionInputFileError(RelatedQuestionError):
    """Exception raised when user enter an invalid question input"""
    """ for data collector """

    def __init__(self, input_file, message):
        self.input_file = input_file
        self.message = message

    def __unicode__(self):
        return (
            f"Invalid input file: {self.input_file}\n{message}"
        )


class FailedToWriteOuputFileError(RelatedQuestionError):
    """Exception raised when program fails to write data to """
    """ output file for data colletor"""

    def __init__(self, output_file, message):
        self.output_file = output_file
        self.message = message

    def __unicode__(self):
        return (
            f"Cannot write to {self.output_file}\n{message}"
        )



#! /usr/bin/env python3
import time
import json
import argparse
import traceback
from collections import OrderedDict
from people_also_ask_it.google import get_simple_answer
from people_also_ask_it.exceptions import (
    InvalidQuestionInputFileError,
    FailedToWriteOuputFileError,
)

def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("--input-file", "-i", help="input file which is a txt file containing list of questions", required=True)
    parser.add_argument("--output-file", "-o", help="output file which is .json file containing a dictionnary of question: answer", required=True)
    parser.add_argument("--proxy-file", "-p", help="proxy file containing list of proxy")

    return parser.parse_args()


def read_questions(input_file):
    try:
        with open(input_file, "r") as fd:
            text = fd.read()
            return OrderedDict.fromkeys(text.strip().split("\n")).keys()
    except Exception:
        message = traceback.format_exc()
        raise InvalidQuestionInputFileError(input_file, message)

def write_question_answers(output_file, data):
    try:
        with open(output_file, "w") as fd:
            fd.write(json.dumps(data))
    except Exception:
        message = traceback.format_exc()
        raise FailedToWriteOuputFileError(output_file, message)


def collect_one_question(question):
    try:
        answer = get_simple_answer(question)
        print(f"{question}: {answer}")
    except Exception:
        traceback.print_exc()
        answer = ""
    return {question: answer}


def collect_data(input_file, output_file, proxy_file=None):
    questions = read_questions(input_file)
    nb_questions = len(questions)
    counter = 0
    data = {}
    start_time = time.time()
    end_time = None
    if proxy_file is None:
        for question in questions:
            counter += 1
            print(f"COLLECTING {counter}/{nb_questions}")
            data.update(collect_one_question(question))
    end_time = time.time()
    collect_time = (end_time - start_time) / 60  #  minutes
    print(
        f"Collected answers for {nb_questions} questions in {collect_time} minutes"
        )
    write_question_answers(output_file, data)

def main():
    args = parse_args()
    collect_data(args.input_file, args.output_file, args.proxy_file)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
from bs4.element import Tag
from bs4 import BeautifulSoup
from operator import attrgetter
from typing import List, Optional
from people_also_ask_it.tools import itemize, tabulate, remove_redundant


FEATURED_SNIPPET_ATTRIBUTES = [
    "response", "heading", "title", "link", "displayed_link",
    "snippet_str", "snippet_data", "date", "snippet_data",
    "snippet_type", "snippet_str_body", "raw_text"
]


def extract_related_questions(document: BeautifulSoup) -> List[str]:
    div_questions = document.find_all("div", class_="related-question-pair")
    get_text = attrgetter("text")
    if not div_questions:
        return []
    questions = list(map(get_text, div_questions))
    return questions


def is_ol_but_not_a_menu(tag):
    return (
        tag.name == "ol"
        and (
            not tag.has_attr("role")
            or (tag.has_attr("role") and tag["role"] != "menu")
            )
        )


def get_tag_heading(tag):
    return (
        tag.find("div", {"role": "heading", "aria-level": "3"})
        or tag.find("div", {"role": "heading"})
    )


def has_youtube_link(tag):
    youtube_links = tag.findAll(
        lambda x: x.name == "a" and "youtube" in x.get("href", "")
    )
    return bool(youtube_links)


def get_raw_text(tag):
    return "\n".join(remove_redundant(tag.strings))


def get_span_text(tag):
    return "\n".join(
            remove_redundant(
                [e.text for e in tag.findAll("span") if e.text]
                )
            )


class FeaturedSnippetParser(object):

    def __init__(self, text: str, tag: Tag):
        self.text = text
        self.tag = tag

    def __getattr__(self, attr):
        if attr in FEATURED_SNIPPET_ATTRIBUTES:
            return None
        raise AttributeError(f'{self.__class__.__name__}.{attr} is invalid.')

    @property
    def raw_text(self):
        return get_raw_text(self.tag)

    def to_dict(self):
        return {
            attr: getattr(self, attr) for attr in FEATURED_SNIPPET_ATTRIBUTES
        }


class SimpleFeaturedSnippetParser(FeaturedSnippetParser):

    @classmethod
    def get_instance(self, text, tag):
        if tag.table is not None:
            return TableFeaturedSnippetParser(text, tag)
        if tag.findAll(is_ol_but_not_a_menu):
            return OrderedFeaturedSnippetParser(text, tag)
        if tag.ul is not None:
            return UnorderedFeaturedSnippetParser(text, tag)
        if get_tag_heading(tag):
            return DefinitionFeaturedSnippetParser(text, tag)
        if has_youtube_link(tag):
            return YoutubeFeaturedSnippetParser(text, tag)

    @property
    def tag_link(self):
        if hasattr(self, "_tag_link"):
            return self._tag_link
        self._tag_link = self.tag.find(
            lambda tag: (
                tag.name == "a"
                and tag.has_attr("href")
                and tag["href"].startswith("http")
                and (tag.h3 or tag.h2) is not None
            )
        )
        return self._tag_link

    @property
    def link(self):
        return self.tag_link["href"] if self.tag_link else None

    @property
    def displayed_link(self):
        return self.tag.cite.text if self.tag.cite else None

    @property
    def title(self):
        if self.tag_link is None:
            return None
        tag_title = self.tag_link.h3 or self.tag_link.h2
        return tag_title.text

    @property
    def heading(self):
        tag_heading = get_tag_heading(self.tag)
        return tag_heading.text

    @property
    def snippet_str(self):
        lines = []
        for field in (
            "heading", "snippet_str_body",
            "displayed_link", "link", "title"
        ):
            if getattr(self, field):
                lines.append(getattr(self, field))
        return "\n".join(lines)

    @property
    def date(self):
        return None

    @property
    def snippet_data(self):
        return None

    @property
    def snippet_type(self):
        return "Unknown Featured Snippet"

    @property
    def snippet_str_body(self):
        return ""


class TableFeaturedSnippetParser(SimpleFeaturedSnippetParser):
    """Example: world university rankings 2019"""

    @property
    def snippet_type(self):
        return "Table Featured Snippet"

    @property
    def snippet_str_body(self):
        header = self.snippet_data["columns"]
        table = self.snippet_data["values"]
        return tabulate(header=header, table=table)

    @property
    def response(self):
        return self.snippet_str_body

    @property
    def snippet_data(self):
        table_tag = self.tag.find("table")
        tr_tags = table_tag.findAll("tr")
        if tr_tags[0].find("th"):
            columns = [
                th_tag.text for th_tag in tr_tags[0].findAll("th")
            ]
            body_table_tags = tr_tags[1:]
        else:
            columns = None
            body_table_tags = tr_tags
        values = [
            [td_tag.text for td_tag in tr_tag.findAll("td")]
            for tr_tag in body_table_tags
        ]
        if columns is None:
            columns = list(range(len(values[0])))
        return {
            "columns": columns,
            "values": values
        }


class OrderedFeaturedSnippetParser(SimpleFeaturedSnippetParser):
    """Example: top grossing movies"""

    @property
    def snippet_type(self):
        return "Ordered Featured Snippet"

    @property
    def response(self):
        return self.snippet_str_body

    @property
    def snippet_str_body(self):
        return "\n".join(itemize(self.snippet_data))

    @property
    def snippet_data(self):
        ol_tags = self.tag.find("ol")
        li_tags = ol_tags.findAll("li")
        return [tag.text for tag in li_tags]


class UnorderedFeaturedSnippetParser(SimpleFeaturedSnippetParser):
    """ What are 3 basic programming languages? """

    @property
    def snippet_type(self):
        return "Unordered Featured Snippet"

    @property
    def snippet_str_body(self):
        return "\n".join(itemize(self.snippet_data))

    @property
    def response(self):
        return self.snippet_str_body

    @property
    def snippet_data(self):
        ul_tag = self.tag.find("ul")
        li_tags = ul_tag.findAll("li")
        return [tag.text for tag in li_tags]


class DefinitionFeaturedSnippetParser(SimpleFeaturedSnippetParser):
    """Why was ho chi minh a hero"""

    @property
    def snippet_type(self):
        return "Definition Featured Snippet"

    @property
    def response(self):
        return self.heading


class YoutubeFeaturedSnippetParser(SimpleFeaturedSnippetParser):
    """Ex: cheetah vs lion"""

    @property
    def snippet_type(self):
        return "Youtube Featured Snippet"

    @property
    def heading(self):
        return ""

    @property
    def response(self):
        return self.link


class MultipleCardsFeaturedSnippetTag(FeaturedSnippetParser):
    """How to make a cold brew coffee"""

    @property
    def heading(self):
        tag_heading = (
            self.tag.find("h3", {"role": "heading"})
            or self.tag.find("h2", {"role": "heading"})
        )
        return tag_heading.text

    @property
    def snippet_type(self):
        return "Multiple Cards Featured Snippet Tag"

    def parse_card(self, tag_card):
        return {
            "heading": tag_card.find("div", {"role": "heading"}).text,
            "title": tag_card.cite.text,
            "link": tag_card.a["href"],
            "raw_text": get_raw_text(tag_card),
        }

    def str_card(self, card_data):
        lines = [card_data["raw_text"]]
        lines.append(f"Link: {card_data['link']}")
        return "\n".join(lines)

    @property
    def snippet_str(self):
        if not self.snippet_data:
            return ""
        return "\n-------------\n".join(map(self.str_card, self.snippet_data))

    @property
    def snippet_data(self):
        return list(map(self.parse_card, self.tag.findAll("g-inner-card")))

    @property
    def response(self):
        return self.snippet_str


class SingleCardFeaturedSnippetParser(FeaturedSnippetParser):
    """What time is it"""

    @property
    def snippet_type(self):
        return "Single Card FeaturedSnippet"

    @property
    def heading(self):
        tag_heading = get_tag_heading(self.tag)
        return get_raw_text(tag_heading)

    @property
    def response(self):
        heading = self.heading
        if heading:
            return heading
        return self.raw_text

    @property
    def raw_text(self):
        return get_span_text(self.tag)


class WholePageTabContainer(FeaturedSnippetParser):
    """Gangnam Style"""

    @property
    def snippet_type(self):
        return "Whole Page Tab Container"

    @property
    def tag_link(self):
        if hasattr(self, "_tag_link"):
            return self._tag_link
        self._tag_link = self.tag.find(
            lambda tag: (
                tag.name == "a"
                and tag.has_attr("href")
                and tag["href"].startswith("http")
                and (tag.h3 or tag.h2) is not None
            )
        )
        return self._tag_link

    @property
    def link(self):
        return self.tag_link["href"] if self.tag_link else None

    @property
    def displayed_link(self):
        return self.tag.cite.text if self.tag.cite else None

    @property
    def title(self):
        if self.tag_link is None:
            return None
        tag_title = self.tag_link.h3 or self.tag_link.h2
        return tag_title.text

    @property
    def response(self):
        return self.raw_text

    @property
    def raw_text(self):
        return get_span_text(self.tag)


def is_simple_featured_snippet_tag(tag):
    class_tuple = tuple(tag.get("class", ""))
    is_xpdopen = (tag.name == "div" and class_tuple == ("xpdopen",))
    if not is_xpdopen:
        return False
    is_xpdopen_of_related_questions = (
        tag.h2 is not None and tag.h2.text == "People also ask"
    )
    return not is_xpdopen_of_related_questions


def is_single_card_featured_snippet_tag(tag):
    is_card_section = (
        tag.name == "div" and "card-section" in tag.get("class", [])
    )
    if not is_card_section:
        return False
    is_card_section_of_tip = tag.text.startswith("Tip:")
    return not is_card_section_of_tip


def is_multiple_card_snippet_tag(tag):
    return (tag.name == "g-section-with-header")


def is_whole_page_tabs_container(tag):
    return (tag.get("id") == "wp-tabs-container")


def is_web_results(tag):
    return (tag.name == "h2" and tag.text == "Web results")


def get_featured_snippet_tag(document):

    def lookup_featured_snippet_tag(tag):
        return (
            is_simple_featured_snippet_tag(tag)
            or is_single_card_featured_snippet_tag(tag)
            or is_multiple_card_snippet_tag(tag)
            or is_web_results(tag)
        )
    whole_page_tag = document.find(is_whole_page_tabs_container)
    tag = document.find(lookup_featured_snippet_tag)
    if tag and is_simple_featured_snippet_tag(tag):
        return tag
    if whole_page_tag:
        return whole_page_tag
    if not tag or tag.name == "h2":
        return None
    return tag


def get_featured_snippet_parser(question, document: BeautifulSoup):
    tag = get_featured_snippet_tag(document)
    if tag is None:
        return
    if is_simple_featured_snippet_tag(tag):
        return SimpleFeaturedSnippetParser.get_instance(question, tag)
    if is_multiple_card_snippet_tag(tag):
        return MultipleCardsFeaturedSnippetTag(question, tag)
    if is_single_card_featured_snippet_tag(tag):
        return SingleCardFeaturedSnippetParser(question, tag)
    if is_whole_page_tabs_container(tag):
        return WholePageTabContainer(question, tag)



# -*- coding: utf-8 -*-
import codecs
import os
import re
import setuptools


def local_file(file):
    return codecs.open(
        os.path.join(os.path.dirname(__file__), file), 'r', 'utf-8'
    )

setuptools.setup(
    name="people_also_ask_it",
    version="0.2",
    author="intelligenzaartificialeitalia.net",
    author_email="",
    packages=setuptools.find_packages(),
    long_description=local_file('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers=[
	"Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Typing :: Typed",
    ],
    install_requires=[
        "beautifulsoup4",
        "requests",
    ],
    python_requires=">=3.6"
)
